import backoff
import requests

from opentelemetry.ext.lightstep.version import __version__


class APIClient:
    """HTTP client to send data to Lightstep"""

    def __init__(
        self, token, url,
    ):
        self._headers = {
            "User-Agent": "otel-ls-python/{}".format(__version__),
            "Accept": "application/octet-stream",
            "Content-Type": "application/octet-stream",
            "Lightstep-Access-Token": token,
        }
        self._url = url

    @backoff.on_exception(backoff.expo, Exception, max_time=5)
    def send(self, content, token=None):
        if token is not None:
            self._headers["Lightstep-Access-Token"] = token

        return requests.post(self._url, headers=self._headers, data=content)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: metrics.proto

import sys

_b = (
    sys.version_info[0] < 3 and (lambda x: x) or (lambda x: x.encode("latin1"))
)
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


import opentelemetry.ext.lightstep.protobuf.collector_pb2 as collector__pb2
from google.protobuf import (
    duration_pb2 as google_dot_protobuf_dot_duration__pb2,
)
from google.protobuf import (
    timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2,
)


DESCRIPTOR = _descriptor.FileDescriptor(
    name="metrics.proto",
    package="lightstep.metrics",
    syntax="proto3",
    serialized_options=_b(
        "\n\031com.lightstep.tracer.grpcP\001Z\tmetricspb\242\002\004LSPB"
    ),
    serialized_pb=_b(
        '\n\rmetrics.proto\x12\x11lightstep.metrics\x1a\x0f\x63ollector.proto\x1a\x1egoogle/protobuf/duration.proto\x1a\x1fgoogle/protobuf/timestamp.proto"\x8f\x02\n\x0bMetricPoint\x12+\n\x04kind\x18\x01 \x01(\x0e\x32\x1d.lightstep.metrics.MetricKind\x12\x13\n\x0bmetric_name\x18\x02 \x01(\t\x12)\n\x05start\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12+\n\x08\x64uration\x18\x04 \x01(\x0b\x32\x19.google.protobuf.Duration\x12-\n\x06labels\x18\x05 \x03(\x0b\x32\x1d.lightstep.collector.KeyValue\x12\x16\n\x0cuint64_value\x18\x06 \x01(\x04H\x00\x12\x16\n\x0c\x64ouble_value\x18\x07 \x01(\x01H\x00\x42\x07\n\x05value"\x89\x01\n\rIngestRequest\x12\x17\n\x0fidempotency_key\x18\x01 \x01(\t\x12/\n\x08reporter\x18\x02 \x01(\x0b\x32\x1d.lightstep.collector.Reporter\x12.\n\x06points\x18\x03 \x03(\x0b\x32\x1e.lightstep.metrics.MetricPoint"\x10\n\x0eIngestResponse*=\n\nMetricKind\x12\x17\n\x13INVALID_METRIC_KIND\x10\x00\x12\x0b\n\x07\x43OUNTER\x10\x01\x12\t\n\x05GAUGE\x10\x02\x42/\n\x19\x63om.lightstep.tracer.grpcP\x01Z\tmetricspb\xa2\x02\x04LSPBb\x06proto3'
    ),
    dependencies=[
        collector__pb2.DESCRIPTOR,
        google_dot_protobuf_dot_duration__pb2.DESCRIPTOR,
        google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,
    ],
)

_METRICKIND = _descriptor.EnumDescriptor(
    name="MetricKind",
    full_name="lightstep.metrics.MetricKind",
    filename=None,
    file=DESCRIPTOR,
    values=[
        _descriptor.EnumValueDescriptor(
            name="INVALID_METRIC_KIND",
            index=0,
            number=0,
            serialized_options=None,
            type=None,
        ),
        _descriptor.EnumValueDescriptor(
            name="COUNTER",
            index=1,
            number=1,
            serialized_options=None,
            type=None,
        ),
        _descriptor.EnumValueDescriptor(
            name="GAUGE", index=2, number=2, serialized_options=None, type=None
        ),
    ],
    containing_type=None,
    serialized_options=None,
    serialized_start=550,
    serialized_end=611,
)
_sym_db.RegisterEnumDescriptor(_METRICKIND)

MetricKind = enum_type_wrapper.EnumTypeWrapper(_METRICKIND)
INVALID_METRIC_KIND = 0
COUNTER = 1
GAUGE = 2


_METRICPOINT = _descriptor.Descriptor(
    name="MetricPoint",
    full_name="lightstep.metrics.MetricPoint",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="kind",
            full_name="lightstep.metrics.MetricPoint.kind",
            index=0,
            number=1,
            type=14,
            cpp_type=8,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="metric_name",
            full_name="lightstep.metrics.MetricPoint.metric_name",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="start",
            full_name="lightstep.metrics.MetricPoint.start",
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="duration",
            full_name="lightstep.metrics.MetricPoint.duration",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="labels",
            full_name="lightstep.metrics.MetricPoint.labels",
            index=4,
            number=5,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="uint64_value",
            full_name="lightstep.metrics.MetricPoint.uint64_value",
            index=5,
            number=6,
            type=4,
            cpp_type=4,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="double_value",
            full_name="lightstep.metrics.MetricPoint.double_value",
            index=6,
            number=7,
            type=1,
            cpp_type=5,
            label=1,
            has_default_value=False,
            default_value=float(0),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="value",
            full_name="lightstep.metrics.MetricPoint.value",
            index=0,
            containing_type=None,
            fields=[],
        ),
    ],
    serialized_start=119,
    serialized_end=390,
)


_INGESTREQUEST = _descriptor.Descriptor(
    name="IngestRequest",
    full_name="lightstep.metrics.IngestRequest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="idempotency_key",
            full_name="lightstep.metrics.IngestRequest.idempotency_key",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="reporter",
            full_name="lightstep.metrics.IngestRequest.reporter",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="points",
            full_name="lightstep.metrics.IngestRequest.points",
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=393,
    serialized_end=530,
)


_INGESTRESPONSE = _descriptor.Descriptor(
    name="IngestResponse",
    full_name="lightstep.metrics.IngestResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=532,
    serialized_end=548,
)

_METRICPOINT.fields_by_name["kind"].enum_type = _METRICKIND
_METRICPOINT.fields_by_name[
    "start"
].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_METRICPOINT.fields_by_name[
    "duration"
].message_type = google_dot_protobuf_dot_duration__pb2._DURATION
_METRICPOINT.fields_by_name["labels"].message_type = collector__pb2._KEYVALUE
_METRICPOINT.oneofs_by_name["value"].fields.append(
    _METRICPOINT.fields_by_name["uint64_value"]
)
_METRICPOINT.fields_by_name[
    "uint64_value"
].containing_oneof = _METRICPOINT.oneofs_by_name["value"]
_METRICPOINT.oneofs_by_name["value"].fields.append(
    _METRICPOINT.fields_by_name["double_value"]
)
_METRICPOINT.fields_by_name[
    "double_value"
].containing_oneof = _METRICPOINT.oneofs_by_name["value"]
_INGESTREQUEST.fields_by_name[
    "reporter"
].message_type = collector__pb2._REPORTER
_INGESTREQUEST.fields_by_name["points"].message_type = _METRICPOINT
DESCRIPTOR.message_types_by_name["MetricPoint"] = _METRICPOINT
DESCRIPTOR.message_types_by_name["IngestRequest"] = _INGESTREQUEST
DESCRIPTOR.message_types_by_name["IngestResponse"] = _INGESTRESPONSE
DESCRIPTOR.enum_types_by_name["MetricKind"] = _METRICKIND
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

MetricPoint = _reflection.GeneratedProtocolMessageType(
    "MetricPoint",
    (_message.Message,),
    {
        "DESCRIPTOR": _METRICPOINT,
        "__module__": "metrics_pb2"
        # @@protoc_insertion_point(class_scope:lightstep.metrics.MetricPoint)
    },
)
_sym_db.RegisterMessage(MetricPoint)

IngestRequest = _reflection.GeneratedProtocolMessageType(
    "IngestRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _INGESTREQUEST,
        "__module__": "metrics_pb2"
        # @@protoc_insertion_point(class_scope:lightstep.metrics.IngestRequest)
    },
)
_sym_db.RegisterMessage(IngestRequest)

IngestResponse = _reflection.GeneratedProtocolMessageType(
    "IngestResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _INGESTRESPONSE,
        "__module__": "metrics_pb2"
        # @@protoc_insertion_point(class_scope:lightstep.metrics.IngestResponse)
    },
)
_sym_db.RegisterMessage(IngestResponse)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



import os
import random
import string

import requests
from google.protobuf.duration_pb2 import Duration
from google.protobuf.timestamp_pb2 import Timestamp

from opentelemetry.ext.lightstep import reporter
from opentelemetry.ext.lightstep.api_client import APIClient
from opentelemetry.ext.lightstep.protobuf.collector_pb2 import KeyValue
from opentelemetry.ext.lightstep.protobuf.metrics_pb2 import (
    IngestRequest,
    MetricKind,
    MetricPoint,
)
from opentelemetry.sdk.metrics.export import (
    MetricRecord,
    MetricsExporter,
    MetricsExportResult,
    Sequence,
)

from .. import util

_MAX_DURATION = 10 * 60  # ten minutes in seconds

METRICS_URL_ENV_VAR = "LS_METRICS_URL"
_DEFAULT_METRICS_URL = os.environ.get(
    METRICS_URL_ENV_VAR, "https://ingest.lightstep.com:443/metrics"
)
_DEFAULT_SERVICE_VERSION = "0.0.0"


class LightstepMetricsExporter(MetricsExporter):
    def _calc_value(self, key, value):
        if self._filters.get(key, MetricKind.GAUGE) == MetricKind.GAUGE:
            return value
        delta = value - self._store.get(key, 0)
        self._store[key] = value
        return delta

    def __init__(
        self,
        name,
        token,
        url=_DEFAULT_METRICS_URL,
        service_version=_DEFAULT_SERVICE_VERSION,
    ):
        self._store = {}
        # only capture metrics used in Lightstep
        self._filters = {
            "runtime.python.gc.count.gen0": MetricKind.GAUGE,
            "runtime.python.gc.count.gen1": MetricKind.GAUGE,
            "runtime.python.gc.count.gen2": MetricKind.GAUGE,
            "runtime.python.cpu.sys": MetricKind.COUNTER,
            "runtime.python.cpu.user": MetricKind.COUNTER,
            "runtime.python.mem.rss": MetricKind.GAUGE,
            "cpu.sys": MetricKind.COUNTER,
            "cpu.user": MetricKind.COUNTER,
            "cpu.total": MetricKind.COUNTER,
            "cpu.usage": MetricKind.COUNTER,
            "mem.available": MetricKind.GAUGE,
            "mem.total": MetricKind.GAUGE,
            "net.bytes_recv": MetricKind.COUNTER,
            "net.bytes_sent": MetricKind.COUNTER,
        }
        self._service_name = name
        self._service_version = service_version
        self._token = token
        self._client = APIClient(token, url=url)
        self._guid = util._generate_guid()
        self._reporter = reporter.get_reporter(
            self._service_name, self._service_version, self._guid
        )
        self._key_length = 30
        self._last_success = 0
        self._labels = [
            KeyValue(key=reporter.HOSTNAME_KEY, string_value=os.uname()[1]),
            KeyValue(
                key=reporter.SERVICE_NAME_KEY, string_value=self._service_name
            ),
            KeyValue(
                key=reporter.SERVICE_VERSION_KEY,
                string_value=self._service_version,
            ),
        ]

    def _ingest_request(self):
        return IngestRequest(
            reporter=self._reporter,
            idempotency_key=self._generate_idempotency_key(),
        )

    def _generate_idempotency_key(self):
        return "".join(
            random.choice(string.ascii_lowercase)
            for i in range(self._key_length)
        )

    def _converted_labels(self, labels):
        # converts labels from otel to ls format
        converted = []
        for key, val in labels:
            converted.append(KeyValue(key=key, string_value=val))
        return converted

    def _should_discard(self, duration):
        # intentionally throw away first report
        return self._last_success == 0 or duration.ToSeconds() > _MAX_DURATION

    def export(
        self, metric_records: Sequence[MetricRecord]
    ) -> MetricsExportResult:
        """Exports a batch of telemetry data.

        Args:
            metric_records: A sequence of `MetricRecord` s. A `MetricRecord`
                contains the metric to be exported, the label set associated
                with that metric, as well as the aggregator used to export the
                current checkpointed value.

        Returns:
            The result of the export
        """
        ingest_request = self._ingest_request()
        start_time = Timestamp()
        start_time.GetCurrentTime()
        duration = Duration()
        duration.FromSeconds(start_time.ToSeconds() - self._last_success)

        for record in metric_records:
            if record.instrument.name not in self._filters.keys():
                continue
            value = 0.0
            if record.aggregator.checkpoint.last is not None:
                value = float(
                    self._calc_value(
                        record.instrument.name,
                        record.aggregator.checkpoint.last,
                    )
                )

            ingest_request.points.add(
                duration=duration,
                start=start_time,
                labels=self._converted_labels(record.labels) + self._labels,
                metric_name=record.instrument.name,
                double_value=value,
                kind=self._filters.get(record.instrument.name),
            )

        if len(ingest_request.points) == 0:
            return MetricsExportResult.SUCCESS

        if self._should_discard(duration):
            self._last_success = start_time.ToSeconds()
            return MetricsExportResult.SUCCESS

        resp = self._client.send(ingest_request.SerializeToString())
        if resp.status_code == requests.codes["ok"]:
            self._last_success = start_time.ToSeconds()
            return MetricsExportResult.SUCCESS
        return MetricsExportResult.FAILURE

    def shutdown(self) -> None:
        """Shuts down the exporter.

        Called when the SDK is shut down.
        """
